
/*
 *  DMOUSE-HANDLER.C
 *
 *  (c)Copyright 1989 by Matthew Dillon, All Rights Reserved
 *
 *  V1.20, last revision 3 August 1989
 *
 *  Note on upping the handler process priority.  This is done to cause the
 *  handler task to get CPU before the current input event completes its
 *  processing so intuition calls made by the process are executed before
 *  the event is propogated.  If said intuition calls block, it's ok
 *  because they are not blocking the input handler process.
 */

#include "dmouse.h"

#ifdef LATTICE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

/*typedef struct Layer	      LAYER;*/
typedef struct IE	    *IEP;
typedef struct IORequest    IORequest;

DMS	*Dms		       = NULL;
IBASE	*IntuitionBase	       = NULL;
GFXBASE *GfxBase	       = NULL;
struct LayersBase  *LayersBase = NULL;
struct ExecBase    *SysBase    = NULL;

static PORT	*IOPort = NULL;     /*	For IPC messages		  */
static LIST	BlankList;	    /*	list of external blanker programs */

static char	STimedout = 0;
static char	MTimedout = 0;
static long	STime = 0, MTime = 0;
#ifdef DEBUG
static long	DBFh = NULL;
#endif


#define IBASE IntuitionBase

NS	Ns = {	0, 0, 64, -1, 1, -1, -1, 0, CUSTOMSCREEN|SCREENQUIET };
IE DummyIE = { 0 };

short	NRMe;	/*  Don't Repeat Mouse Events   */

#ifdef LATTICE
__asm __saveds IEP handler(register __a0 IE *event);
#else
IE *handler();
#endif

LAYER *WhichMouseLayer ARGS((void));
void sendrequest ARGS((long, IE *));
LAYER *WhichMouseLayer ARGS((void));
int doipcmsg ARGS((short));
void DeleteBlanker ARGS((IORequest *));
__saveds void noname ARGS((void));

#ifdef LATTICE
__saveds void
noname()
#else
_main()
#endif
{

    reg DMS *dms;
    IOR  *ior;
    INT addhand;

#ifndef LATTICE
    geta4();
#endif
    SysBase = *(struct ExecBase **)4;
#ifdef LATTICE
    {
	DOSBase = (struct DosLibrary *)OpenLibrary("dos.library", 0);
    }
#endif
    {
	reg PROC *proc = (PROC *)FindTask(NULL);
	proc->pr_ConsoleTask = NULL;
    }
    NRMe = 0;
    dms = Dms = (DMS *)FindPort(PORTNAME);
    if (!dms)
	return;
    dms->Port.mp_Flags = PA_SIGNAL;
    dms->Port.mp_SigBit = AllocSignal(-1);
    dms->Port.mp_SigTask = FindTask(NULL);
    dms->HandTask = dms->Port.mp_SigTask;
    ior = CreateStdIO(&dms->Port);
    IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 0)

    GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", 0);
    LayersBase = (struct LayersBase *)OpenLibrary("layers.library", 0);
    IOPort = CreatePort("DMouse.ipc", 0);
    NewList(&BlankList);

    if (!IntuitionBase || !GfxBase || !LayersBase)
	goto startupfail;
    addhand.is_Node.ln_Pri = dms->IPri;
    addhand.is_Code = (FPTR)handler;
    addhand.is_Data = NULL;

    if (OpenDevice("input.device", 0, ior, 0)) {
	goto startupfail;
    } else {
	SCR *scr = NULL;
	uword *SprSavePtr = NULL;
	long ipc_mask;

	Signal(dms->ShakeTask, 1 << dms->ShakeSig);
	ior->io_Command = IND_ADDHANDLER;
	ior->io_Data = (APTR)&addhand;
	ior->io_Message.mn_Node.ln_Type = NT_MESSAGE;
	DoIO(ior);

	ipc_mask = 1 << IOPort->mp_SigBit;

	for (;;) {
	    reg long sigs = Wait(SBF_C|(1<<dms->Port.mp_SigBit)|ipc_mask);
	    if (sigs & (1 << dms->Port.mp_SigBit)) {
		reg REQ *msg;
		while (msg = (REQ *)GetMsg(&dms->Port)) {
		    switch((long)msg->Msg.mn_Node.ln_Name) {
		    case REQ_SCREENON:
			if (scr)
			    CloseScreen(scr);
			scr = NULL;
			doipcmsg(0x82);
			break;
		    case REQ_SCREENOFF:
			if (scr)
			    ScreenToFront(scr);
			if (doipcmsg(0x83) == 0 && scr == NULL) {
			    if (scr = OpenScreen(&Ns))
				SetRGB4(&scr->ViewPort, 0, 0, 0, 0);
			}
			break;
		    case REQ_MOUSEON:
			if (SprSavePtr) {
			    register COPINIT *ci = GfxBase->copinit;
			    ci->sprstrtup[1] = (ulong)SprSavePtr >> 16;
			    ci->sprstrtup[3] = (uword)(long)SprSavePtr;
			    SprSavePtr = NULL;
			}
			doipcmsg(0x80);
			break;
		    case REQ_MOUSEOFF:
			if (doipcmsg(0x81) == 0) {
			    reg COPINIT *ci = GfxBase->copinit;
			    if (ci->sprstrtup[0] != 0x120 || ci->sprstrtup[2] != 0x122)
				break;

			    if (!SprSavePtr)
				SprSavePtr = (uword *)((ci->sprstrtup[1] << 16) | ci->sprstrtup[3]);
			    ci->sprstrtup[1] = (ulong)dms->NoSprData >> 16;
			    ci->sprstrtup[3] = (uword)(long)dms->NoSprData;
			}
			break;
		    case REQ_DOCMD:
			{
			    long fh = (long)Open("nil:", 1006);
			    Execute(dms->Cmd, NULL, fh);
			    if (fh)
				Close(fh);
			}
			break;
		    case REQ_RAWMOUSE:
			{
			    register LAYER *layer;

			    NRMe = 0;
			    Forbid();
			    layer = WhichMouseLayer();
			    if (msg->ie_Code == IECODE_RBUTTON && dms->LMBEnable && (msg->ie_Qualifi
r & dms->RQual)) {
				register WIN *win;
				if (layer && (win = (WIN *)layer->Window) && !(win->Flags & BACKDROP) && (w
n->NextWindow || win->WScreen->FirstWindow != win)) {
				    if (dms->Workbench)
					WindowToBack(win);
				    else
					BehindLayer(0, layer);
				} else if (IBASE->FirstScreen)
				    ScreenToBack(IBASE->FirstScreen);
			    }
			    if (layer && layer->Window) {
				if (msg->ie_Code == IECODE_LBUTTON && !(((WIN *)layer->Window)->Flags & BAC
DROP) && dms->LMBEnable && layer->ClipRect && layer->ClipRect->Next) {
				    /*
				     *	Note: Case where it is the 'first' click in a series, where dms->CTi
e is
				     *	      garbage, works properly no matter what DoubleClick returns.
				     */
				    if (dms->LQual == 0 || (msg->ie_Qualifier & dms->LQual)) {
					if ((APTR)dms->CWin == layer->Window && DoubleClick(dms->CTime.tv_secs, dm
->CTime.tv_micro, msg->ie_TimeStamp.tv_secs, msg->ie_TimeStamp.tv_micro))
					    --dms->CLeft;
					else
					    dms->CLeft = dms->Clicks - 1;
					dms->CTime = msg->ie_TimeStamp;
					dms->CWin = (WIN *)layer->Window;
					if (dms->CLeft == 0) {
					    dms->CLeft = dms->Clicks;
					    if (dms->Workbench)
						WindowToFront((WIN *)layer->Window);
					    else
						UpfrontLayer(0, layer);
					}
				    }
				}
				if ((dms->AAEnable & 1) && (void *)layer->Window != (void *)IBASE->ActiveWi
dow && msg->ie_Code == IECODE_NOBUTTON && !(msg->ie_Qualifier & 0x7000)) {
				    if (!IBASE->ActiveWindow || !IBASE->ActiveWindow->FirstRequest)
					ActivateWindow((WIN *)layer->Window);
				}
			    }
			    Permit();
			}
			break;
		    case REQ_RAWKEY:
			{
			    register LAYER *layer;

			    Forbid();
			    layer = WhichMouseLayer();
			    if (layer && layer->Window && (void *)layer->Window != (void *)IBASE->Ac
iveWindow) {
				if (!IBASE->ActiveWindow || !IBASE->ActiveWindow->FirstRequest)
				    ActivateWindow((WIN *)layer->Window);
			    }
			    Permit();
			}
			break;
#ifdef DEBUG
		    case REQ_DEBUG:
			{
			    char buf[128];
			    if (!DBFh) {
				DBFh = Open("con:0/0/400/100/dmouse-debug", 1006);
				if (!DBFh)
				    break;
			    }
			    sprintf(buf, "%02lx %04lx %04lx\n",
				msg->ie_Class,
				msg->ie_Code,
				msg->ie_Qualifier
			    );
			    Write(DBFh, buf, strlen(buf));
			}
			break;
		    case REQ_DEBUGOFF:
			if (DBFh) {
			    Close(DBFh);
			    DBFh = NULL;
			}
			break;
#endif
		    }
		    FreeMem(msg, msg->Msg.mn_Length);
		}
	    }
	    if (sigs & SBF_C)
		break;

	    /*
	     *	IPC request.
	     */

	    if (sigs & ipc_mask) {
		reg IORequest *ior;
		while (ior = (IORequest *)GetMsg(IOPort)) {
		    long req = 0;

		    if (ior->io_Message.mn_Node.ln_Type == NT_REPLYMSG) {
			FreeMem(ior, ior->io_Message.mn_Length);
			continue;
		    }

		    ior->io_Error = 0;
		    switch(ior->io_Command) {
		    case 0x80:	/* mouse on  */
			req = REQ_MOUSEON;
			break;
		    case 0x81:	/* mouse off */
			req = REQ_MOUSEOFF;
			MTimedout = 1;
			break;
		    case 0x82:	/* screen on */
			req = REQ_SCREENON;
			break;
		    case 0x83:	/* screen off*/
			req = REQ_SCREENOFF;
			STimedout = 1;
			break;
		    case 0x84:	/* add hand  */
			AddHead(&BlankList, ior);
			ior = NULL;
			break;
		    case 0x85:	/* rem hand  */
			DeleteBlanker(ior);
			ior = NULL;
			break;
		    }
		    if (req)
			sendrequest(req, NULL);
		    if (ior)
			ReplyMsg(&ior->io_Message);
		}
	    }
	}
#ifdef DEBUG
	if (DBFh) {
	    Close(DBFh);
	    DBFh = NULL;
	}
#endif
	ior->io_Command = IND_REMHANDLER;
	ior->io_Data = (APTR)&addhand;
	ior->io_Message.mn_Node.ln_Type = NT_MESSAGE;
	DoIO(ior);
	ior->io_Command = IND_WRITEEVENT;	/*  NULL EVENT	*/
	ior->io_Length = sizeof(IE);
	ior->io_Data = (APTR)&DummyIE;
	ior->io_Message.mn_Node.ln_Type = NT_MESSAGE;
	DoIO(ior);
	CloseDevice(ior);
	{
	    reg MSG *msg;
	    while (msg = GetMsg(&dms->Port))
		FreeMem(msg, msg->mn_Length);
	}
	if (scr)
	    CloseScreen(scr);
	if (SprSavePtr) {
	    reg COPINIT *ci = GfxBase->copinit;
	    ci->sprstrtup[1] = (ulong)SprSavePtr >> 16;
	    ci->sprstrtup[3] = (uword)(long)SprSavePtr;
	    SprSavePtr = NULL;
	}
    }
    goto closedown;
startupfail:
    dms->StartupError = 1;
    Signal(dms->ShakeTask, 1 << dms->ShakeSig);
    Wait(SBF_C);
closedown:
    DeleteStdIO(ior);
fail:
    if (IOPort) {
	IORequest *ior;     /*	wait for RemReq messages */

	doipcmsg(0x86);     /*  send closedown requests  */
	Forbid();
	while (GetHead(&BlankList)) {
	    WaitPort(IOPort);
	    while (ior = (IORequest *)GetMsg(IOPort)) {
		if (ior->io_Message.mn_Node.ln_Type == NT_REPLYMSG) {
		    FreeMem(ior, ior->io_Message.mn_Length);
		    continue;
		}
		if (ior->io_Command == 0x85)    /*  receive remove req  */
		    DeleteBlanker(ior);
		else
		    ReplyMsg(&ior->io_Message);              /*  ignore other reqs   */
	    }
	}
	DeletePort(IOPort);
	Permit();
    }
    if (IntuitionBase)
	CloseLibrary((LIB *)IntuitionBase);
    if (GfxBase)
	CloseLibrary((LIB *)GfxBase);
    if (LayersBase)
	CloseLibrary((LIB *)LayersBase);
#ifdef LATTICE
    {
	CloseLibrary((LIB *)DOSBase);
    }
#endif
    Forbid();
    Signal(dms->ShakeTask, 1 << dms->ShakeSig);
}

void
DeleteBlanker(ior)
IORequest *ior;
{
    IORequest *io2;

    ior->io_Error = 0;
    for (io2 = GetHead(&BlankList); io2; io2 = GetSucc(io2)) {
	if (io2->io_Unit == ior->io_Unit) {
	    Remove(io2);
	    if (ior)
		ReplyMsg(&ior->io_Message);
	    ReplyMsg(&io2->io_Message);
	    ior = NULL;
	}
    }
    if (ior) {
	ior->io_Error = -1;
	ReplyMsg(&ior->io_Message);
    }
}

doipcmsg(cmd)
short cmd;
{
    short count = 0;
    short flags = 1 << (cmd & 0x7F);    /*  enable flags */
    IORequest *iob, *io;

    for (iob = GetHead(&BlankList); iob; iob = GetSucc(iob)) {
	if (cmd == 0x86 || (iob->io_Flags & flags)) {
	    io = AllocMem(sizeof(IORequest), MEMF_PUBLIC|MEMF_CLEAR);
	    if (io) {
		io->io_Command = cmd;
		io->io_Unit = iob->io_Unit;
		io->io_Message.mn_ReplyPort = IOPort;
		io->io_Message.mn_Length = sizeof(IORequest);
		PutMsg(iob->io_Message.mn_ReplyPort, &io->io_Message);
		++count;
	    }
	}
    }
    return((int)count);
}

/*
 *  The INPUT.DEVICE HANDLER
 *
 *  Note that for Lattice V5.02 and beyond we need no
 *  assembly tags because we can specify that the arguments
 *  are passed in registers.
 */

#ifndef LATTICE

#asm
	    ;	A0 = pointer to event linked list
	    ;	A1 = pointer to my data segment
	    ;	return new event linked list in D0

	    public  _CHandler

_handler:
	    movem.l D2/D3/A0/A1/A4/A6,-(sp)
	    jsr     _CHandler
	    movem.l (sp)+,D2/D3/A0/A1/A4/A6
	    rts

#endasm

#endif

/*
 *  (1) Accellerate mouse movements.
 *  (2) Auto-Select window
 */

#ifdef LATTICE

__asm __saveds IEP
handler(register __a0 IE *Ev)
{

#else

IE *
CHandler(scr0, scr1, Ev)
long scr0;
long scr1;
IE *Ev;
{
#endif
    reg IE *ev;
    reg DMS *dms;

#ifndef LATTICE
    geta4();
#endif
    dms = Dms;
    for (ev = Ev; ev; ev = Ev->ie_NextEvent) {
#ifdef DEBUG
	if (dms->Debug) {
	    if (ev->ie_Class != IECLASS_TIMER)
		sendrequest(REQ_DEBUG, ev);
	} else if (DBFh) {
	    sendrequest(REQ_DEBUGOFF, ev);
	}
#endif
	switch(ev->ie_Class) {
	case IECLASS_RAWMOUSE:
	    /*
	     *	Mouse events restore both the screen and mouse pointer.
	     */

	    STime = ev->ie_TimeStamp.tv_secs + dms->STo;
	    MTime = ev->ie_TimeStamp.tv_secs + dms->MTo;
	    if (STimedout)
		sendrequest(REQ_SCREENON, ev);
	    if (MTimedout)
		sendrequest(REQ_MOUSEON, ev);
	    STimedout = MTimedout = 0;

	    /*
	     *	Mouse Acceleration
	     */
	    {
		register short n;
		register short s;

		if (dms->Acc != 1) {
		    n = ev->ie_X;
		    s = 1;
		    if (n < 0) {
			n = -n;
			s = -1;
		    }
		    if (n > dms->AThresh)
			ev->ie_X = s * (short)((n - dms->AThresh - 1) * dms->Acc + dms->AThresh + 1)

		    n = ev->ie_Y;
		    s = 1;
		    if (n < 0) {
			n = -n;
			s = -1;
		    }
		    if (n > dms->AThresh)
			ev->ie_Y = s * (short)((n - dms->AThresh - 1) * dms->Acc + dms->AThresh + 1)

		}
	    }

	    /*
	     *	Auto Activate and LMB (win/scrn front/bak)
	     */

	    if (dms->LMBEnable && ev->ie_Code == IECODE_RBUTTON && (ev->ie_Qualifier &
dms->RQual))
		ev->ie_Class = IECLASS_NULL;	/*  remove event    */
	    if (NRMe == 0 && ((dms->AAEnable & 1) || dms->LMBEnable)) {
		register short old;
		NRMe = 1;
		if (ev->ie_Code != IECODE_NOBUTTON)
		    old = SetTaskPri(dms->Port.mp_SigTask, 21);
		sendrequest(REQ_RAWMOUSE, ev);
		if (ev->ie_Code != IECODE_NOBUTTON)
		    SetTaskPri(dms->Port.mp_SigTask, old);
	    }
	    break;
	case IECLASS_RAWKEY:
	    /*
	     *	Keyboard events will kill the screen timeout but not
	     *	the mouse timeout.  Note that the priority of the
	     *	co-process must be upped to ensure it is able to make the
	     *	window active before the keystroke is passed further.
	     *
	     *	key releases are ignored
	     *
	     *	note: ie_Qualifier may or may not have bit 15 set
	     */
	    if (ev->ie_Code & 0x80)
		break;
	    if (dms->AAEnable & 2) {
		register short old = SetTaskPri(dms->Port.mp_SigTask, 21);
		sendrequest(REQ_RAWKEY, ev);
		SetTaskPri(dms->Port.mp_SigTask, old);
	    }
	    STime = ev->ie_TimeStamp.tv_secs + dms->STo;
	    if (STimedout) {
		sendrequest(REQ_SCREENON, ev);
		if (dms->MTo == 0)
		    sendrequest(REQ_MOUSEON, ev);
	    }
	    STimedout = 0;

	    if (ev->ie_Code == dms->Code && (ev->ie_Qualifier | 0x8000) == dms->Qual) 

		sendrequest(REQ_DOCMD, ev);
		ev->ie_Class = IECLASS_NULL;	/*  remove event    */
	    }
	    break;
	case IECLASS_TIMER:
	    /*
	     *	On a timer event, if timeout has occured execute the operation
	     *	and reset the timeout.	Note that this will cause continuous
	     *	timeouts every STo and MTo seconds... required because at any
	     *	time Intuition might turn the mouse back on or open a screen or
	     *	something and I want the blanker's to work in the long run.
	     */
	    {
		register long old;
		if (dms->Reset) {
		    dms->Reset = 0;
		    STime = ev->ie_TimeStamp.tv_secs + dms->STo;
		    MTime = ev->ie_TimeStamp.tv_secs + dms->MTo;
		}
		if (dms->STo && (old = STime - ev->ie_TimeStamp.tv_secs) < 0) {
		    STime = ev->ie_TimeStamp.tv_secs + dms->STo + 10;
		    STimedout = 1;
		    MTimedout = 1;
		    if (old > -10) {
			sendrequest(REQ_SCREENOFF, ev);
			sendrequest(REQ_MOUSEOFF, ev);
		    }
		}
		if (dms->MTo && (old = MTime - ev->ie_TimeStamp.tv_secs) < 0) {
		    MTime = ev->ie_TimeStamp.tv_secs + dms->MTo + 1;
		    MTimedout = 1;
		    if (old > -10)
			sendrequest(REQ_MOUSEOFF, ev);
		}
	    }
	    break;
	}
    }
    return(Ev);
}

void
sendrequest(creq, ev)
long creq;
reg IE *ev;
{
    reg REQ *req = AllocMem(sizeof(REQ), MEMF_PUBLIC);

    if (req) {
	req->Msg.mn_Node.ln_Name = (char *)creq;
	req->Msg.mn_ReplyPort = NULL;
	req->Msg.mn_Length = sizeof(REQ);
	if (ev) {
	    req->ie_Class= ev->ie_Class;
	    req->ie_Code = ev->ie_Code;
	    req->ie_Qualifier = ev->ie_Qualifier;
	    req->ie_TimeStamp = ev->ie_TimeStamp;
	}
	PutMsg(&Dms->Port, (MSG *)req);
    }
}

LAYER *
WhichMouseLayer()
{
    register struct IntuitionBase *ib = IBASE;
    register LAYER *layer = NULL;
    register SCR *scr = ib->FirstScreen;

    for (scr = ib->FirstScreen; scr; scr = scr->NextScreen) {
	register short mousey = ib->MouseY;
	register short mousex = ib->MouseX;
	if (!(scr->ViewPort.Modes & LACE))
	    mousey >>= 1;
	if (!(scr->ViewPort.Modes & HIRES))
	    mousex >>= 1;
	if (layer = WhichLayer(&scr->LayerInfo, mousex, mousey - scr->ViewPort.DyOffse
))
	    break;
	if (mousey >= scr->ViewPort.DyOffset)
	    break;
    }
    return(layer);
}


